;    awmessage.pbi
;    Version 0.5 [2012/04/28]
;    Copyright (C) 2012 Ronny Krueger
;
;    This file is part of AWPB-Tools.
;
;    AWPB-Tools is free software: you can redistribute it and/or modify
;    it under the terms of the GNU Lesser General Public License as published by
;    the Free Software Foundation, either version 3 of the License, or
;    (at your option) any later version.
;
;    AWPB-Tools is distributed in the hope that it will be useful,
;    but WITHOUT ANY WARRANTY; without even the implied warranty of
;    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;    GNU Lesser General Public License for more details.
;
;    You should have received a copy of the GNU Lesser General Public License
;    along with AWPB-Tools.  If not, see <http://www.gnu.org/licenses/>.

XIncludeFile "awsupport.pbi"
XIncludeFile "awchunkyfile.pbi"
XIncludeFile "awdate.pbi"
XIncludeFile "awcode.pbi"
XIncludeFile "awcharset.pbi"

;{ Constants
; Priorities
#AWPB_MSG_PRIO_NONE       = 0             ;* keine Prioritt
#AWPB_MSG_PRIO_HIGHEST    = 1             ;* hchste Prioritt
#AWPB_MSG_PRIO_HIGH       = 2             ;* hohe Prioritt
#AWPB_MSG_PRIO_NORMAL     = 3             ;* normale Prioritt
#AWPB_MSG_PRIO_LOW        = 4             ;* niedrige Prioritt
#AWPB_MSG_PRIO_LOWEST     = 5             ;* niedrigste Prioritt

; Content-Types
#AWPB_MSG_TYPE_AUTO       = 0
#AWPB_MSG_TYPE_PLAIN      = 1
#AWPB_MSG_TYPE_HTML       = 2
#AWPB_MSG_MP_RELATED      = 3             ;* multipart/related als Content-Type
#AWPB_MSG_MP_MIXED        = 4             ;* multipart/mixed als Content-Type
#AWPB_MSG_MP_ALTERNATIVE  = 5             ;* multipart/alternative als Content-Type
#AWPB_MSG_TYPE_BINARY     = 6
#AWPB_MSG_TYPE_ATTACHMENT = 7
#AWPB_MSG_TYPE_INLINE     = 8

; Encodings
#AWPB_MSG_ENCODE_AUTO     = 0
#AWPB_MSG_ENCODE_QP       = 1
#AWPB_MSG_ENCODE_B64      = 2

; Header-Fields
#AWPB_HDR_SUBJECT                   = 1
#AWPB_HDR_FROM                      = 2
#AWPB_HDR_CC                        = 3
#AWPB_HDR_CC_MAILS                  = 4
#AWPB_HDR_TO                        = 5
#AWPB_HDR_TO_MAILS                  = 6
#AWPB_HDR_BCC                       = 7
#AWPB_HDR_BCC_MAILS                 = 8
#AWPB_HDR_RETURNPATH                = 9
#AWPB_HDR_MESSAGEID                 = 10
#AWPB_HDR_DATE                      = 11
#AWPB_HDR_DISPOSITIONNOTIFICATIONTO = 12
#AWPB_HDR_REPLYTO                   = 13
#AWPB_HDR_CONTENTTYPE               = 14
#AWPB_HDR_CONTENTSUBTYPE            = 15
#AWPB_HDR_CONTENTCHARSET            = 16
#AWPB_HDR_CONTENTTRANSFERENCODING   = 17
#AWPB_HDR_BOUNDARY                  = 18
#AWPB_HDR_CONTENTDISPOSITION        = 19
#AWPB_HDR_CONTENTNAME               = 20
#AWPB_HDR_CONTENTFILENAME           = 21
#AWPB_HDR_CONTENTID                 = 22
#AWPB_HDR_XPRIORITY                 = 23
#AWPB_HDR_FULLCONTENTTYPE           = 24

;}
;{ structures
Structure AWPB_MSG_HeaderField             ; HeaderField
  key.s                                   ; Key, also das was vorm : bezw vom = steht
  value.s                                 ; der jeweilige Eintrag, max bis zum ;
EndStructure
Structure AWPB_MSG_Header
  Map dat.AWPB_MSG_HeaderField()           ; Eine Headerzeile = eine Map von Eintrgen
EndStructure
Structure AWPB_MSG_MimePart
  Map  Header.AWPB_MSG_Header()            ; Message Header (alle Headerzeilen dieses Parts)
  List MimePart.AWPB_MSG_MimePart()       ; Message SubParts (sofern multipart, alle subparts darin)
  body.i                                  ; Pointer to BodyData, RAW - including all CR/LF
  bodylen.i                               ; Length of BodyData in bytes (including all CR/LF)
  justptr.b                               ; if false FreeMem, if true body is part of msg-src
EndStructure
Structure AWPB_MSG_Msg Extends AWPB_MSG_MimePart
  src.i                                   ; source, ascii-format, 1 byte / char, with CR/LF
  bcc.s                                   ; BCC Empfngerliste
  decoded.b                               ; Flag, #True wenn dekodiert, sonst #False
EndStructure
Structure AWPB_MSG_AttList
  *Next.AWPB_MSG_AttList
  *Att.AWPB_MSG_MimePart
EndStructure
;}
;{ support functions
Procedure.s ____MSG_GenBoundary()                                                         ; Boundary-Generator fr Multipart-Mails
  Protected dt.i, bnd.s
  
  ; die Boundary erstellen
  ; Format: Boundary_YYYMDDhhmmss_RND_
  ; YYY Jahr, M Monat, DD Tag, hh Stunde, mm Minute, ss Sekunde - jeweils in HEX-Format mit entsprechender Zeichenanzahl
  ; RND zufllige 3-Stellige HEX-Zahl) zwischen 000 und FFF
  ; dadurch wird gesichert, das in einer Mail, die mehrere Parts enthlt, jeder Part eine eigene Boundary erhlt
  
  dt  = Date()
  bnd = "Boundary_" + RSet(Hex(Year(dt)), 3, "0") + RSet(Hex(Month(dt)), 1) + RSet(Hex(Day(dt)), 2, "0") + RSet(Hex(Hour(dt)), 2, "0") + RSet(Hex(Minute(dt)), 2, "0") + RSet(Hex(Second(dt)), 2, "0") + "_" + RSet(Hex(Random($FFF)), 3, "0") + "_"
  
  ProcedureReturn bnd
EndProcedure
Procedure.b ____MSG_NeedQuotes(txt.s)                                   ; Test, ob der bergebene String in "" eingeschlossen werden muss
  Protected c.c, *a, i.i, l.i
  
  *a = @txt
  l  = Len(txt)
  If (PeekC(*a) = 34) And (PeekC(*a + l - 1) = 34)
    ; already surrounded by quotes
    ProcedureReturn #False
  EndIf
  For i = 1 To l
    c = PeekC(*a) : *a + SizeOf(c)
    If Not (((c >= 'A') And (c <= 'Z')) Or ((c >= 'a') And (c <= 'z')) Or ((c >= '0') And (c <= '9')))
      ProcedureReturn #True
    EndIf
  Next
  
  ProcedureReturn #False
EndProcedure
Procedure.s ____MSG_AddQuotesIfNeeded(txt.s)                            ; falls der bergebene String in "" eingeschlossen werden muss dieses tun und das ergebnis zurckgeben
  If ____MSG_NeedQuotes(txt)
    txt = Chr(34) + txt + Chr(34)
  EndIf
  ProcedureReturn txt
EndProcedure    
Procedure.i ____MSG_CountCharsToEncode(text.s, encodews.b = #True)      ; Anzahl der Zeichen, die QP-Encoded werden mssen ermitteln
  Protected *b, i.i, l.i, c.c, cnt.i
  
  l   = Len(text)
  *b  = @text
  cnt = 0
  For i = 1 To l
    c = PeekC(*b) : *b + 1
    If (c < 33) Or (c > 126) Or (c = 61) Or (c = '!') Or (c = '"') Or (c = '#') Or (c = '$') Or (c = '@') Or (c = '[') Or (c = '\') Or (c = ']') Or (c = '^') Or (c = '`') Or (c = '{') Or (c = '|') Or (c = '}') Or (c = '~')
      If (c = 9) Or (c = 32)
        If (PeekC(*b) = $0d) Or (PeekC(*b - 2) = $0a) Or encodews
          cnt + 1
        EndIf           
      Else
        If (c <> 10) And (c <> 13)
          cnt + 1
        EndIf
      EndIf
    EndIf
  Next i
  
  ProcedureReturn cnt
EndProcedure  
Procedure.i ____MSG_MaxCharCountForQPLineLen(text.s, linelen.i, startpos.i, encodews.b = #True)   ; maximale Anzahl der Zeichen ermitteln, die in eine Zeile codiert werden knnen
  Protected *b, len.i, c.c, cnt.i, tl.i
  
  tl  = Len(text)
  cnt = 0
  If startpos < tl
    *b  = @text + ((startpos - 1) * SizeOf(c))
    len = 0
    tl  - startpos + 1
    While tl > 0
      c = PeekC(*b)
      If (c < 33) Or (c > 126) Or (c = 61) Or (c = '!') Or (c = '"') Or (c = '#') Or (c = '$') Or (c = '@') Or (c = '[') Or (c = '\') Or (c = ']') Or (c = '^') Or (c = '`') Or (c = '{') Or (c = '|') Or (c = '}') Or (c = '~')
        ; encode char
        If (c = 9) Or (c = 32)            ; need only be encoded before/after CRLF
          If (PeekC(*b) = $0d) Or (PeekC(*b - 2) = $0a) Or encodews
            len + 3
          Else
            len + 1
          EndIf           
        Else
          If (c = 10) Or (c = 13)
            len + 1
          Else
            len + 3
          EndIf
        EndIf
      Else
        len + 1
      EndIf
      If len < linelen : cnt + 1 : Else : Break : EndIf
      *b + SizeOf(Character)
      tl - 1
    Wend
  EndIf
  ProcedureReturn cnt
EndProcedure  
Procedure.b ____MSG_SingleDataExpected(hdrname.s)         ; Header testen, ob mehrere Daten (mit ; getrennte key=value-Paare) erwartet werden (#False) oder nicht (#True)
  ;- probably incomplete
  If hdrname = "CONTENT-TYPE"
    ProcedureReturn #False
  ElseIf hdrname = "CONTENT-DISPOSITION"
    ProcedureReturn #False
    
  EndIf
  ProcedureReturn #True
EndProcedure
Procedure.b ____MSG_NeedEncoding(text.s, ignoretab.b = #True, ignorespc.b = #True)            ; Test, ob der Text kodiert werden muss, also ob er fr 7-Bit eMails ungltige Zeichen enthlt
  Protected *a, i.i, c.c
  
  *a  = @text
  For i = 1 To Len(text)
    c = PeekC(*a)
    If ignoretab And ignorespc
      If (c > 126) Or ((c < 32) And (c <> 9)) Or (c = 34)
        ProcedureReturn #True
      EndIf
    ElseIf ignoretab And Not ignorespc
      If (c > 126) Or ((c < 33) And (c <> 9)) Or (c = 34)
        ProcedureReturn #True
      EndIf
    ElseIf Not ignoretab And ignorespc
      If (c > 126) Or (c < 32) Or (c = 34)
        ProcedureReturn #True
      EndIf
    Else
      If (c > 126) Or (c < 33) Or (c = 34)
        ProcedureReturn #True
      EndIf
    EndIf
    *a + SizeOf(c)
  Next
  
  ProcedureReturn #False
EndProcedure
Procedure.s ____MSG_EncodeInline(text.s, charset.s = "", encode.i = #AWPB_MSG_ENCODE_AUTO, ignoretab.b = #False, ignorespc.b = #False) ;- TODO    ; Inline-Encoder; erzeugt =?...?.?...?= Strings sofern ntig
  Protected res.s, enclen.i, *encbin, a.i, f.s, tlen.i, lines.i, llen.i, codelen.i, pos.i, txt.s, fl.b
  If encode = #AWPB_MSG_ENCODE_AUTO
    If ____MSG_NeedEncoding(text, ignoretab, ignorespc)
      encode = #AWPB_MSG_ENCODE_QP
    EndIf
  EndIf
  
  If Not charset : charset = ____CS_GetSysCharset() : EndIf
  codelen = Len(charset) + 7
  pos     = 1
  tlen    = Len(text)
  res     = ""
  fl      = #True
  
  While pos <= tlen
    Select encode
      Case #AWPB_MSG_ENCODE_AUTO
        llen = 60
        If llen + pos > tlen
          txt = Mid(text, pos)
          pos = tlen + 1
        Else
          txt = Mid(text, pos, llen)
          pos + llen
        EndIf
      Case #AWPB_MSG_ENCODE_B64
        llen = 40 - codelen
        If llen + pos > tlen
          txt = Mid(text, pos)
          pos = tlen + 1
        Else
          txt = Mid(text, pos, llen)
          pos + llen
        EndIf
        *encbin = ____CODE_EncodeBase64Str(txt, Len(txt), @enclen, #PB_Any)
        If *encbin
          txt = "=?" + charset + "?B?" + ____SUP_TrimWS(PeekS(*encbin, enclen, #PB_Ascii)) + "?="
          FreeMemory(*encbin)
        EndIf
      Case #AWPB_MSG_ENCODE_QP
        llen = ____MSG_MaxCharCountForQPLineLen(text, 60 - codelen, pos)
        If llen
          txt  = Mid(text, pos, llen)
          pos  + llen
          *encbin = ____CODE_EncodeQuotedPrintableStr(txt, Len(txt), @enclen, #PB_Any, #True)
          If *encbin
            txt = "=?" + charset + "?Q?" + ____SUP_TrimWS(ReplaceString(PeekS(*encbin, enclen, #PB_Ascii), Chr(32), "_")) + "?="
            FreeMemory(*encbin)
          EndIf
        Else
          txt  = Mid(text, pos)
        EndIf
    EndSelect
    If fl
      fl = #False
      res + txt
    Else
      res + Chr(13) + Chr(10) + Chr(9) + txt
    EndIf
  Wend
  ProcedureReturn res
EndProcedure
Procedure.s ____MSG_DecodeInline(txt.s)                                                   ; Decoder fr Inline-Encodierten ( =?charset?c?daten?= )
  Protected charset.s, codec.s, p1.i, p2.i, p3.i, p4.i, buf.s, blen.i, tmp.s
  
  ; wird nach jedem dekodierten block rekursiv aufgerufen, da mehr als ein block enthalten sein kann
  ; =?CHARSET?CODIERUNG?TEXT?=
  ; Codierung = B fr Base64 oder Q fr Quoted Printable
  If txt
    p1 = FindString(txt, "=?", 1)                           ; =? suchen          
    If p1                                                   ; gefunden, string enthlt kodierten bereich
      p2   = FindString(txt, "?", p1 + 2)                   ; ? suchen
      If p2                                                 ; gefunden
;       charset = Mid(txt, p1 + 1, p2 - p1 - 1)             ;- TODO Charset auslesen (wird aktuell noch nicht benutzt)
        p3      = FindString(txt, "?", p2 + 1)              ; ? suchen
        If p3                                               ; gefunden 
          codec = UCase(Mid(txt, p2 + 1, p3 - p2 - 1))      ; Codec auslesen
          p4    = FindString(txt, "?=", p3 + 1)             ; Ende des kodierten Textes suchen
          If p4                                             ; gefunden
            blen = p4 - p3 - 1                              ; Lnge des Textes
            If codec = "B"
              buf  = Space(blen)
              PokeS(@buf, Mid(txt, p3 + 1, blen), blen, #PB_Ascii)    ; Text als ASCII auslesen und dekodieren (Base64)
              ProcedureReturn ____CODE_DecodeBase64Str(@buf, blen) + ____MSG_DecodeInline(Mid(txt, p4 + 2))      ; dekodierten Text mit dem dekodierten Rest der Zeile zurckgeben
            ElseIf codec = "Q"              
              buf  = Space(blen)
              PokeS(@buf, ReplaceString(Mid(txt, p3 + 1, blen), "_", " "), blen, #PB_Ascii)   ; Text als ASCII auslesen und dekodieren (QuotedPrintable)
              ProcedureReturn  ____CODE_DecodeQuotedPrintableStr(@buf, blen) + ____MSG_DecodeInline(Mid(txt, p4 + 2))    ; dekodierten Text mit dem dekodierten Rest der Zeile zurckgeben
            Else
              ProcedureReturn Mid(txt, p3 + 1, blen) + ____MSG_DecodeInline(Mid(txt, p4 + 2))    ; unbekannter codec, Text mit dem dekodierten Rest der Zeile zurckgeben
            EndIf
          EndIf
        EndIf
      EndIf
    EndIf
  EndIf
  
  ProcedureReturn txt
EndProcedure
;}
;{ Message-Object
Interface AWPB_Message
  ; public
  Destroy                 ()
  Clear                   ()
  Decode.b                ()
  BuildAttachmentList.i   (*part.AWPB_MSG_MimePart = #Null)
  FreeAttachmentList      (*AttList.AWPB_MSG_AttList)
  Encode.b                ()
  SetBodyBinary.b         (*src, srclen.i, name.s, filename.s, id.s = "", *part.AWPB_MSG_MimePart = #Null)
  SetBodyText.b           (body.s, type.i, charset.s = "", encode.i = #AWPB_MSG_ENCODE_AUTO, *part.AWPB_MSG_MimePart = #Null)
  SetMultipart            (type.i, *part.AWPB_MSG_MimePart = #Null)
  AddMultipart.i          (type.i, *part.AWPB_MSG_MimePart = #Null)
  AddPartHTML.b           (html.s, charset.s = "", encode.i = #AWPB_MSG_ENCODE_AUTO, *part.AWPB_MSG_MimePart = #Null)
  AddPartText.b           (text.s, charset.s = "", encode.i = #AWPB_MSG_ENCODE_AUTO, *part.AWPB_MSG_MimePart = #Null)
  AddPartBinary.b         (*src, srclen.i, name.s, filename.s, id.s = "", *part.AWPB_MSG_MimePart = #Null)
  AddPartBinaryFile.b     (name.s, filename.s, id.s = "", *part.AWPB_MSG_MimePart = #Null)
  SetMailSource.b         (src.s)
  SetMailSourceBin.b      (*src, srclen.i)
  GetMailSource.s         ()
  LoadEML.b               (filename.s)
  SaveEML.b               (filename.s)
  Load.b                  (filename.s)
  Save.b                  (filename.s, complevel.i = #AWPB_ZIP_COMP_MAX)
  GetHeaderField.s        (field.i, *part.AWPB_MSG_MimePart = #Null)
  SetHeaderField          (field.i, value.s, *part.AWPB_MSG_MimePart = #Null)
  SetContentType          (contenttype.s, contentsubtype.s, *part.AWPB_MSG_MimePart = #Null)
  FindPartCID.i           (cid.s, casesensitive.b = #False, *part.AWPB_MSG_MimePart = #Null)
  FindPartName.i          (name.s, casesensitive.b = #False, *part.AWPB_MSG_MimePart = #Null)
  FindPartFileName.i      (filename.s, casesensitive.b = #False, *part.AWPB_MSG_MimePart = #Null)
  GetDate.i               (*part.AWPB_MSG_MimePart = #Null)
  SetDate                 (datevalue.i, *part.AWPB_MSG_MimePart = #Null)
  SetDateNow              (*part.AWPB_MSG_MimePart = #Null)
  GetXPriority.i          (*part.AWPB_MSG_MimePart = #Null)
  SetXPriority            (priority.i, *part.AWPB_MSG_MimePart = #Null)
  GetHTML.s               ()
  GetPlain.s              ()
  GetPartAsString.s       (*part.AWPB_MSG_MimePart)
  GetPartAsBin.i          (*part.AWPB_MSG_MimePart, *partlen)
  SavePartToFile.b        (*part.AWPB_MSG_MimePart, filename.s)
  GetPlainAsHTML.s        ()
  GetSize.i               ()
  
  ; private
  _ClearPart              (*msgpart.AWPB_MSG_MimePart)
  _GetHeaderValue.s       (*msgpart.AWPB_MSG_MimePart, Key1.s, Key2.s)
  _GetHeaderData.s        (*msgpart.AWPB_MSG_MimePart, Key.s)
  _AnalyzeMimePart.b      (*msgpart.AWPB_MSG_MimePart, *src, maxlen.i)
  _AddHeader              (*msgpart.AWPB_MSG_MimePart, hdr.s, value.s)
  _AddHeaderValue         (*msgpart.AWPB_MSG_MimePart, hdr.s, key.s, value.s)
  _BuildHeader.s          (*msgpart.AWPB_MSG_MimePart)
  _CalcHeaderSize.i       (*msgpart.AWPB_MSG_MimePart)
  _CalcSize.i             (*msgpart.AWPB_MSG_MimePart)
  _BuildPart.i            (*msgpart.AWPB_MSG_MimePart, *partlen, addbufsize.i)
  _GetHTMLPart.i          (*part.AWPB_MSG_MimePart)
  _GetPlainPart.i         (*part.AWPB_MSG_MimePart)
  _GetMailSourceBin.i     ()
  _IsDecoded.b            ()
EndInterface
Structure ____AWPB_Message
  ; private
    *vt.AWPB_Message
    *msg.AWPB_MSG_Msg
  ; public
    
EndStructure
;{ Member Procedures
Procedure.i AWPB_CreateMessageObj()
  Protected *this.____AWPB_Message
  
  *this = AllocateMemory(SizeOf(____AWPB_Message))
  If *this 
    *this\vt  = ?____AWPB_Message_VT
    *this\msg = AllocateMemory(SizeOf(AWPB_MSG_Msg))
    If *this\msg
      InitializeStructure(*this\msg, AWPB_MSG_Msg)
      *this\msg\bcc     = ""
      *this\msg\decoded = #False      
      *this\msg\src     = #Null
      
      ProcedureReturn *this
    EndIf
  EndIf
  
  ProcedureReturn #Null
EndProcedure
Procedure   ____AWPB_Message_ClearPart(*this.____AWPB_Message, *msgpart.AWPB_MSG_MimePart)                                  ; Messagestruktur (ein Part + seine subparts) bereinigen
  Protected bl.i
  
  If *msgpart
    ForEach *msgpart\Header()
      ForEach *msgpart\Header()\dat()
        *msgpart\Header()\dat()\key   = ""
        *msgpart\Header()\dat()\value = ""        
      Next
      ClearMap(*msgpart\Header()\dat())
    Next
    ClearMap(*msgpart\Header())   
    ForEach *msgpart\MimePart()
      ____AWPB_Message_ClearPart(*this, *msgpart\MimePart())
    Next
    ClearList(*msgpart\MimePart())
    If Not *msgpart\justptr 
      If *msgpart\body
        FreeMemory(*msgpart\body)
        *msgpart\body = #Null
      EndIf
    EndIf
    *msgpart\bodylen = 0    
  EndIf
EndProcedure
Procedure   ____AWPB_Message_Clear(*this.____AWPB_Message)
  ____AWPB_Message_ClearPart(*this, *this\msg)
  If *this\msg\src : FreeMemory(*this\msg\src) : *this\msg\src = #Null : EndIf
  *this\msg\bcc     = ""    
  *this\msg\decoded = #False
EndProcedure
Procedure.s ____AWPB_Message_GetHeaderValue(*this.____AWPB_Message, *msgpart.AWPB_MSG_MimePart, Key1.s, Key2.s)
  Protected res.s
  
  res = ""
  If Not *msgpart : *msgpart = *this\msg : EndIf
  If FindMapElement(*msgpart\Header(), Key1)
    If FindMapElement(*msgpart\Header()\dat(), Key2)
      res = *msgpart\Header()\dat()\value
    EndIf
  EndIf
  
  ProcedureReturn res
EndProcedure
Procedure.s ____AWPB_Message_GetHeaderData(*this.____AWPB_Message, *msgpart.AWPB_MSG_MimePart, Key.s)
  ProcedureReturn ____AWPB_Message_GetHeaderValue(*this, *msgpart, Key, Key)
EndProcedure
Procedure.b ____AWPB_Message_AnalyzeMimePart(*this.____AWPB_Message, *msgpart.AWPB_MSG_MimePart, *src, maxlen.i)            ; Message-Sub-Part zerlegen
  Protected *start, *p, *ende, *lstart, *lende, *dstart, *dende, *d, *ep, *dstart2, *bodystart, *bodyend, *boundary, *bstart, *bende, *safe
  Protected len.i, c.a, pos.i, llen.i, tmp.i, blen.i
  Protected hdrname.s, hdrdatal.s, hdrdatar.s, boundary.s, test.c, firstbound.b, lastbound.b, res.b, t.l, *m.AWPB_MSG_Header, uhdrname.s, uhdrdatal.s
  
  ; Header ermitteln
  ; -> CRLFCRLF suchen (Leerzeile)
  res      = #False
  *start   = *src                                      ; *start == start des headers
  len      = -1
  *bodyend = *src + maxlen - 1
  *ende    = #Null
  For pos = 1 To maxlen
    If PeekL(*src) = $0a0d0a0d
      len   = *src - *start + 2                       ; len    == lnge des headers (inkl. einem CRLF)
      *ende = *src + 1                                ; *ende  == letztes Zeichen des headers
      *src  + 4                                       ; *src   == start des nchsten Blockes nach der Leerzeile
      Break     
    EndIf
    *src + 1
  Next pos
  *bodystart = *src
  
  ; Header parsen
  *lstart = *start                                    ; *lstart == anfang der zeile
  For *p = *start To *ende
    If PeekW(*p) = $0a0d
      *p + 2                                          ; *p      == start der nchsten zeile
      c = PeekA(*p)
      If (c <> 9) And (c <> 32)                       ; KEIN zeilenumbruch
        *lende = *p - 1                               ; *lende  == letztes zeichen der zeile
        llen   = *lende - *lstart + 1                 ; llen    == lnge der zeile
        For *dstart = *lstart To *lende
          If PeekA(*dstart) = Asc(":")
            *dstart + 1
            Break
          EndIf
        Next *dstart
        hdrname  = ____SUP_TrimWS(PeekS(*lstart, *dstart - *lstart - 1, #PB_Ascii))
        uhdrname = UCase(hdrname)
        *dstart + 1
        With *msgpart\header(uhdrname)
          If ____MSG_SingleDataExpected(uhdrname)             ; keine unterteilung
            hdrdatar             = ____SUP_TrimWS(PeekS(*dstart, *lende - *dstart, #PB_Ascii))
            \dat(uhdrname)\key   = hdrname
            \dat(uhdrname)\value = hdrdatar
          Else                                          ; parameter sammeln
            *ep      = #Null
            *dstart2 = *dstart
            For *d = *dstart To *lende 
              c = PeekA(*d)
              If c = 34
                Repeat : *d + 1 : Until PeekA(*d) = 34
              ElseIf (c = Asc("=")) And (*ep = #Null)
                *ep = *d
              ElseIf (c = Asc(";")) Or ((c = $0d) And ((*ep <> #Null) Or (*d = *lende - 1)))
                If *ep                                ; = gefunden
                  hdrdatal              = ____SUP_TrimWS(PeekS(*dstart2, *ep - *dstart2, #PB_Ascii))
                  uhdrdatal             = UCase(hdrdatal)
                  hdrdatar              = ____SUP_TrimWSQ(PeekS(*ep + 1, *d - *ep - 1, #PB_Ascii))
                  *ep                   = #Null
                  \dat(uhdrdatal)\key   = hdrdatal
                  \dat(uhdrdatal)\value = hdrdatar
                Else
                  hdrdatar              = ____SUP_TrimWSQ(PeekS(*dstart, *d - *dstart2, #PB_Ascii))
                  \dat(uhdrname)\key    = hdrname
                  \dat(uhdrname)\value  = hdrdatar
                EndIf
                *d       + 1 
                *dstart2 = *d
              EndIf
            Next *d
          EndIf
        EndWith
        *lstart = *p                                  ; *lstart == anfang der nchsten zeile
      EndIf       
    EndIf
  Next *p
  
  boundary = ____AWPB_Message_GetHeaderValue(*this, *msgpart, "CONTENT-TYPE", "BOUNDARY")    
  If boundary ; erste boundary suchen
    blen       = Len(boundary)
    *boundary  = AllocateMemory(blen + 1)
    If *boundary = #Null : ProcedureReturn #False : EndIf
    PokeS(*boundary, boundary, blen, #PB_Ascii)   
    *bstart = *bodystart
    For *d = *bodystart To *bodyend
      If PeekW(*d) = $2d2d
        *bende = *d - 1
        If CompareMemoryString(*d + 2, *boundary, 0, blen, #PB_Ascii) = #PB_String_Equal
          *safe  = *d
          *d     + 2 
          *d     + blen
          If PeekW(*d) = $2d2d        ; das war die letzte boundary
            If PeekW(*d + 2) = $0a0d  ; und der zeilenumbruch danach; wenn nicht, kann es nicht die boundary sein
              *msgpart\body    = *bstart
              *msgpart\bodylen = *bende - *bstart + 2
              *msgpart\justptr = #True
              *bstart      = *d + 4
              res          = #True
              Break
            Else
              *d = *safe
            EndIf
          ElseIf PeekW(*d) = $0a0d    ; normaler Zeilenumbruch; wenn nicht, kann es nicht die boundary sein
            *msgpart\body    = *bstart
            *msgpart\bodylen = *bende - *bstart + 2             
            *msgpart\justptr = #True
            *bstart      = *d + 2
            res          = #True
            Break
          Else
            *d = *safe
          EndIf
        EndIf
      EndIf
    Next *d
    *bodystart = *bstart  ; *bodystart == erstes zeichen nach der boundary
    
    For *d = *bodystart To *bodyend
      If PeekW(*d) = $2d2d
        *bende = *d - 1
        If CompareMemoryString(*d + 2, *boundary, 0, blen, #PB_Ascii) = #PB_String_Equal
          *safe  = *d
          *d     + 2 
          *d     + blen
          If PeekW(*d) = $2d2d        ; das war die letzte boundary
            If PeekW(*d + 2) = $0a0d  ; und der zeilenumbruch danach; wenn nicht, kann es nicht die boundary sein
              AddElement(*msgpart\MimePart())
              res = ____AWPB_Message_AnalyzeMimePart(*this, @*msgpart\MimePart(), *bstart, *bende - *bstart + 2)
              *bstart      = *d + 4
            Else
              *d = *safe
            EndIf
          ElseIf PeekW(*d) = $0a0d    ; normaler Zeilenumbruch; wenn nicht, kann es nicht die boundary sein
            AddElement(*msgpart\MimePart())
            res = ____AWPB_Message_AnalyzeMimePart(*this, @*msgpart\MimePart(), *bstart, *bende - *bstart + 2)              
            *bstart = *d + 2
          Else
            *d = *safe
          EndIf
        EndIf
      EndIf
    Next *d
  Else
    ; keine boundary, rest als body speichern
    *msgpart\body    = *bodystart
    *msgpart\bodylen = *bodyend - *bodystart - 2  
    *msgpart\justptr = #True
  EndIf
  
  ProcedureReturn #True
EndProcedure
Procedure.b ____AWPB_Message_Decode(*this.____AWPB_Message)                                                                 ; Message zerlegen
  *this\msg\decoded = ____AWPB_Message_AnalyzeMimePart(*this, *this\msg, *this\msg\src, MemorySize(*this\msg\src))  
  ProcedureReturn *this\msg\decoded
EndProcedure
Procedure.i ____AWPB_Message_BuildAttachmentList(*this.____AWPB_Message, *part.AWPB_MSG_MimePart)                           ; eine Liste aller speicherbaren Attachments erstellen (Inline-Sachen wie z.B. bilder in gmx-werbungen werden hier nicht eingefgt)
  Protected cd.s, *al.AWPB_MSG_AttList, *al2.AWPB_MSG_AttList, *al3.AWPB_MSG_AttList
  If Not *part : *part = *this\msg : EndIf
  cd  = ____AWPB_Message_GetHeaderData(*this, *part, "CONTENT-DISPOSITION")  
  *al = #Null
  If UCase(cd) = "ATTACHMENT" 
    *al = AllocateMemory(SizeOf(AWPB_MSG_AttList))
    If *al
      *al\Att  = *part
      *al\Next = #Null
    Else
      ProcedureReturn #Null  
    EndIf
  Else
    ForEach *part\MimePart()
      *al2 = ____AWPB_Message_BuildAttachmentList(*this, *part\MimePart())
      If *al2
        If *al
          *al3 = *al
          While *al3\Next : *al3 = *al3\Next : Wend
          *al3\Next = *al2
        Else
          *al = *al2
        EndIf
      EndIf
    Next
  EndIf
  
  ProcedureReturn *al
EndProcedure
Procedure   ____AWPB_Message_FreeAttachmentList(*this.____AWPB_Message, *AttList.AWPB_MSG_AttList)
  Protected *al.AWPB_MSG_AttList
  
  While *AttList
    *al = *AttList\Next
    FreeMemory(*AttList)
    *AttList = *al
  Wend
EndProcedure
Procedure   ____AWPB_Message_AddHeader(*this.____AWPB_Message, *msgpart.AWPB_MSG_MimePart, hdr.s, value.s)                  ; Header hinzufgen hdr=ABC value=GHI -> "ABC: GHI"
  Protected uhdr.s
  
  If Not *msgpart : *msgpart = *this\msg : EndIf
  uhdr                                  = UCase(hdr)
  *msgpart\Header(uhdr)\dat(uhdr)\key   = hdr
  *msgpart\Header(uhdr)\dat(uhdr)\value = value
EndProcedure
Procedure   ____AWPB_Message_AddHeaderValue(*this.____AWPB_Message, *msgpart.AWPB_MSG_MimePart, hdr.s, key.s, value.s)      ; Header hinzufgen hdr=ABC key=DEF value=GHI -> "ABC: DEF=GHI;"
  Protected uhdr.s, ukey.s
  
  If Not *msgpart : *msgpart = *this\msg : EndIf
  uhdr                                  = UCase(hdr)
  ukey                                  = UCase(key)
  *msgpart\Header(uhdr)\dat(ukey)\key   = key
  *msgpart\Header(uhdr)\dat(ukey)\value = value
EndProcedure
Procedure.s ____AWPB_Message_BuildHeader(*this.____AWPB_Message, *msgpart.AWPB_MSG_MimePart)                                ; fgt eine Header-MAP zu einem Headerstring zusammen
  Protected hdr.s, key.s, subkey.s, dat.s, subhdr.s, usubkey.s, ukey.s
  
  hdr = ""
  If Not *msgpart : *msgpart = *this\msg : EndIf
  ForEach *msgpart\Header()
    ukey   = MapKey(*msgpart\Header())
    key    = *msgpart\Header()\dat(ukey)\key
    subhdr = ""
    ForEach *msgpart\Header()\dat()
      usubkey = MapKey(*msgpart\Header()\dat())
      subkey  = *msgpart\Header()\dat()\key
      dat     = *msgpart\Header()\dat()\value
      If usubkey = ukey
        If subhdr
          subhdr = dat + ";" + Chr(13) + Chr(10) + Chr(9) + subhdr
        Else
          subhdr = dat
        EndIf
      Else
        If subhdr
          subhdr + ";" + Chr(13) + Chr(10) + Chr(9)
        EndIf
        subhdr + subkey + "=" + dat
      EndIf
    Next
    If hdr : hdr + Chr(13) + Chr(10) : EndIf        
    hdr + key + ": " + subhdr
  Next
  
  ProcedureReturn hdr
EndProcedure
Procedure.i ____AWPB_Message_CalcHeaderSize(*this.____AWPB_Message, *msgpart.AWPB_MSG_MimePart)                             ; Berechnet die Gre des Headers in Zeichenanzahl
  Protected key.i, dat.i, subkey.i, ukey.s, usubkey.s, subhdr.i, hdr.i
  
  hdr = 0
  If Not *msgpart : *msgpart = *this\msg : EndIf
  ForEach *msgpart\Header()
    ukey   = MapKey(*msgpart\Header())
    key    = Len(*msgpart\Header()\dat(ukey)\key)
    subhdr = 0
    ForEach *msgpart\Header()\dat()
      usubkey = MapKey(*msgpart\Header()\dat())
      subkey  = Len(*msgpart\Header()\dat()\key)
      dat     = Len(*msgpart\Header()\dat()\value)
      If usubkey = ukey
        If subhdr
          subhdr + dat + 4
        Else
          subhdr = dat
        EndIf
      Else
        If subhdr
          subhdr + 4
        EndIf
        subhdr + subkey + 1 + dat
      EndIf
    Next
    If hdr : hdr + 2 : EndIf
    hdr + key + 2 + subhdr
  Next
  
  ProcedureReturn hdr
EndProcedure
Procedure.i ____AWPB_Message_CalcSize(*this.____AWPB_Message, *msgpart.AWPB_MSG_MimePart)                                   ; Berechnet die Gre der encodierten Message in Bytes, inkl. Header und allen Steuercodes
  Protected res.i, bound.s, bl.i
  
  If Not *msgpart : *msgpart = *this\msg : EndIf
  bound = ____AWPB_Message_GetHeaderValue(*this, *msgpart, "CONTENT-TYPE", "BOUNDARY")
  res   + ____AWPB_Message_CalcHeaderSize(*this, *msgpart)
  If *msgpart\body
    res + *msgpart\bodylen + 4
  Else
    res + 2
  EndIf
  If bound
    bl  = Len(bound)
    res + 4 + bl
    ForEach *msgpart\MimePart()
      res + ____AWPB_Message_CalcSize(*this, *msgpart\MimePart()) + 6 + bl
    Next
    res + 4
  EndIf
  
  ProcedureReturn res
EndProcedure
Procedure.i ____AWPB_Message_BuildPart(*this.____AWPB_Message, *msgpart.AWPB_MSG_MimePart, *partlen, addbufsize.i)          ; Generiert einen Messagepart inkl. Header und Body
  Protected *src, newlen.i, header.s, *partsrc, bound.s, *srcpos, bl.i, hl.i, plen.i
  
  header = ""
  bound  = ""
  plen   = 0
  newlen = ____AWPB_Message_CalcSize(*this, *msgpart) + addbufsize
  If newlen
    *src = AllocateMemory(newlen)   
    If *src
      bound   = ____AWPB_Message_GetHeaderValue(*this, *msgpart, "CONTENT-TYPE", "BOUNDARY")
      *srcpos = *src
      header  = ____AWPB_Message_BuildHeader(*this, *msgpart)
      bl      = Len(bound)
      hl      = Len(header)
      PokeS(*srcpos, header, hl, #PB_Ascii)          : *srcpos + hl
      If *msgpart\body
        PokeL(*srcpos, $0a0d0a0d)                    : *srcpos + 4
        CopyMemory(*msgpart\body, *srcpos, *msgpart\bodylen) : *srcpos + *msgpart\bodylen
      Else
        PokeW(*srcpos, $0a0d)                        : *srcpos + 2
      EndIf     
      If bound
        PokeL(*srcpos, $2d2d0a0d)                    : *srcpos + 4
        PokeS(*srcpos, bound, bl, #PB_Ascii)         : *srcpos + bl
        ForEach *msgpart\MimePart()
          *partsrc = ____AWPB_Message_BuildPart(*this, *msgpart\MimePart(), @plen, 0)
          If *partsrc
            PokeW(*srcpos, $0a0d)                    : *srcpos + 2
            CopyMemory(*partsrc, *srcpos, plen)      : *srcpos + plen
            PokeL(*srcpos, $2d2d0a0d)                : *srcpos + 4
            PokeS(*srcpos, bound, bl, #PB_Ascii)     : *srcpos + bl
            FreeMemory(*partsrc)
          EndIf
        Next
        PokeL(*srcpos, $0a0d2d2d)                    : *srcpos + 4
      EndIf
      plen = *srcpos - *src
    EndIf
  EndIf
  If *src
    PokeL(*partlen, plen)
  Else
    PokeL(*partlen, 0)
  EndIf
  ProcedureReturn *src
EndProcedure
Procedure.b ____AWPB_Message_Encode(*this.____AWPB_Message)                                                                 ; Generiert die Message inkl CRLF Abschluss
  Protected newlen.i, *src
  
  *src = ____AWPB_Message_BuildPart(*this, *this\msg, @newlen, 3)
  If *src
    PokeW(*src + newlen, $0a0d) : newlen + 2
    PokeA(*src + newlen, 0)     : newlen + 1
    *this\msg\src = *src
    ProcedureReturn #True
  EndIf
  ProcedureReturn #False
EndProcedure
Procedure.b ____AWPB_Message_SetBodyBinary(*this.____AWPB_Message, *src, srclen.i, name.s, filename.s, id.s, *part.AWPB_MSG_MimePart)          ; setzt den Body eines Parts mit Binrdaten (diese werden automatisch Base64 encodiert)
  Protected len.i
  
  If Not *part : *part = *this\msg : EndIf
  If *src And srclen
    ____AWPB_Message_AddHeader(*this, *part, "Content-Type", "application/force-download")
    ____AWPB_Message_AddHeader(*this, *part, "Content-Transfer-Encoding", "base64")
    If name     : ____AWPB_Message_AddHeaderValue(*this, *part, "Content-Type", "name", ____MSG_AddQuotesIfNeeded(name))                : EndIf
    If filename : ____AWPB_Message_AddHeaderValue(*this, *part, "Content-Disposition", "filename", ____MSG_AddQuotesIfNeeded(filename)) : EndIf
    If id 
      ____AWPB_Message_AddHeader(*this, *part, "Content-Disposition", "inline")
      ____AWPB_Message_AddHeader(*this, *part, "Content-ID", "<"+id+">")
    Else
      ____AWPB_Message_AddHeader(*this, *part, "Content-Disposition", "attachment") 
    EndIf
    
    *part\body    = ____CODE_EncodeBase64(*src, srclen, @len)
    *part\bodylen = len
    *part\justptr = #False
    
    If *part\bodylen : ProcedureReturn #True : EndIf  
  EndIf 
  ProcedureReturn #False
EndProcedure
Procedure.b ____AWPB_Message_SetBodyText(*this.____AWPB_Message, body.s, type.i, charset.s, encode.i, *part.AWPB_MSG_MimePart)                 ; setzt den Body eines Parts als ASCII-String, der ntigenfalls/erzwungen codiert wird
  Protected len.i

  If Not *part : *part = *this\msg : EndIf
  Select type
    Case #AWPB_MSG_TYPE_PLAIN :   ____AWPB_Message_AddHeader(*this, *part, "Content-Type", "text/plain")
    Case #AWPB_MSG_TYPE_HTML  :   ____AWPB_Message_AddHeader(*this, *part, "Content-Type", "text/html")
    Default                   :   ____AWPB_Message_AddHeader(*this, *part, "Content-Type", "text/plain")
  EndSelect
  
  If Not charset : charset = ____CS_GetSysCharset() : EndIf
  
  ____AWPB_Message_AddHeaderValue(*this, *part, "Content-Type", "charset", ____MSG_AddQuotesIfNeeded(charset))      
      
  If encode = #AWPB_MSG_ENCODE_AUTO
    If ____MSG_NeedEncoding(body)
      encode = #AWPB_MSG_ENCODE_QP
    EndIf
  EndIf
    
  Select encode
    Case #AWPB_MSG_ENCODE_B64 
      ____AWPB_Message_AddHeader(*this, *part, "Content-Transfer-Encoding", "base64")
      *part\body    = ____CODE_EncodeBase64Str(body, Len(body), @len)
      *part\bodylen = len
      *part\justptr = #False
    Case #AWPB_MSG_ENCODE_QP
      ____AWPB_Message_AddHeader(*this, *part, "Content-Transfer-Encoding", "quoted-printable")
      *part\body    = ____CODE_EncodeQuotedPrintableStr(body, Len(body), @len)
      *part\bodylen = len
      *part\justptr = #False
    Default
      ____AWPB_Message_AddHeader(*this, *part, "Content-Transfer-Encoding", "8bit")
      len           = Len(body)
      *part\body    = AllocateMemory(len + 4)
      If *part\body
        *part\bodylen = len + 2
        PokeS(*part\body, body, len, #PB_Ascii)
        PokeW(*part\body + len, $0a0d)
      Else
        *part\bodylen = 0
      EndIf
      *part\justptr = #False
  EndSelect
  If *part\bodylen : ProcedureReturn #True : EndIf  
  ProcedureReturn #False
EndProcedure
Procedure   ____AWPB_Message_SetMultipart(*this.____AWPB_Message, type.i, *part.AWPB_MSG_MimePart)                                             ; setzt den Multipart-Header + Boundary eines Parts
  If Not *part : *part = *this\msg : EndIf
  Select type
    Case #AWPB_MSG_MP_ALTERNATIVE  : ____AWPB_Message_AddHeader(*this, *part, "Content-Type", "multipart/alternative")
    Case #AWPB_MSG_MP_RELATED      : ____AWPB_Message_AddHeader(*this, *part, "Content-Type", "multipart/related")
    Default                        : ____AWPB_Message_AddHeader(*this, *part, "Content-Type", "multipart/mixed")
  EndSelect   
  ____AWPB_Message_AddHeaderValue(*this, *part, "Content-Type", "boundary", ____MSG_GenBoundary())
  *part\body    = #Null
  *part\bodylen = 0
  *part\justptr = #False
EndProcedure
Procedure.i ____AWPB_Message_AddMultipart(*this.____AWPB_Message, type.i, *part.AWPB_MSG_MimePart)                                             ; hngt einen neuen Part an die Mail an und setzt den Multipart-Header + Boundary
  Protected *msgpart.AWPB_MSG_MimePart
  
  *msgpart = #Null
  If Not *part : *part = *this\msg : EndIf
  AddElement(*part\MimePart())
  *msgpart = *part\MimePart()
  If *msgpart
    ____AWPB_Message_SetMultipart(*this, *msgpart, type)
    ProcedureReturn *msgpart
  EndIf
  
  ProcedureReturn #Null
EndProcedure
Procedure.b ____AWPB_Message_AddPartHTML(*this.____AWPB_Message, html.s, charset.s, encode.i, *part.AWPB_MSG_MimePart)                         ; fgt einen html-part der Mail hinzu und setzt gleich den (ggf codierten) Body
  Protected *msgpart.AWPB_MSG_MimePart
  
  *msgpart = #Null
  If Not *part : *part = *this\msg : EndIf
  AddElement(*part\MimePart())
  *msgpart = *part\MimePart()
  If *msgpart
    If charset = "" : charset = ____CS_GetHTMLCharset(html) : EndIf
    If ____AWPB_Message_SetBodyText(*this, html, #AWPB_MSG_TYPE_HTML, charset, encode, *msgpart)
      ProcedureReturn #True
    EndIf
    DeleteElement(*part\MimePart())
  EndIf
  
  ProcedureReturn #False
EndProcedure
Procedure.b ____AWPB_Message_AddPartText(*this.____AWPB_Message, text.s, charset.s, encode.i, *part.AWPB_MSG_MimePart)                         ; fgt einen PlainText-part der Mail hinzu und setzt gleich den (ggf codierten) Body
  Protected *msgpart.AWPB_MSG_MimePart
  
  *msgpart = #Null
  If Not *part : *part = *this\msg : EndIf
  AddElement(*part\MimePart())
  *msgpart = *part\MimePart()
  If *msgpart
    If ____AWPB_Message_SetBodyText(*this, text, #AWPB_MSG_TYPE_PLAIN, charset, encode, *msgpart)
      ProcedureReturn #True
    EndIf
    DeleteElement(*part\MimePart())
  EndIf
  
  ProcedureReturn #False
EndProcedure
Procedure.b ____AWPB_Message_AddPartBinary(*this.____AWPB_Message, *src, srclen.i, name.s, filename.s, id.s, *part.AWPB_MSG_MimePart)  ; fgt einen attachment-part der Mail hinzu und setzt gleich den base64 codierten Body
  Protected *msgpart.AWPB_MSG_MimePart
  
  *msgpart = #Null
  If Not *part : *part = *this\msg : EndIf
  AddElement(*part\MimePart())
  *msgpart = *part\MimePart()
  If *msgpart
    If ____AWPB_Message_SetBodyBinary(*this, *src, srclen, name, GetFilePart(filename), id, *msgpart)
      ProcedureReturn #True
    EndIf
    DeleteElement(*part\MimePart())
  EndIf
  
  ProcedureReturn #False
EndProcedure
Procedure.b ____AWPB_Message_AddPartBinaryFile(*this.____AWPB_Message, name.s, filename.s, id.s, *part.AWPB_MSG_MimePart)              ; fgt einen attachment-part aus einer datei der Mail hinzu und setzt gleich den base64 codierten Body
  Protected file.i, length.i, bytes.i, *txt, res.b
  
  res = #False
  If Not *part : *part = *this\msg : EndIf
  file     = ReadFile(#PB_Any, filename)
  If IsFile(file)
    length = Lof(file)
    If length > 0
      *txt = AllocateMemory(length)
      If *txt
        bytes = ReadData(file, *txt, length)
        If bytes = length
          res = ____AWPB_Message_AddPartBinary(*this, *txt, length, name, filename, id, *part)
        EndIf
        FreeMemory(*txt)
      EndIf
    EndIf
    CloseFile(file)
  EndIf   
  
  ProcedureReturn res
EndProcedure
Procedure.b ____AWPB_Message_SetMailSource(*this.____AWPB_Message, src.s)                                                   ; Mailquellcode setzen; es wird ein String bergeben
  Protected len.i
  
  ____AWPB_Message_Clear(*this)
  If src 
    len               = Len(src)
    *this\msg\decoded = #False
    *this\msg\src     = AllocateMemory(len + 1)
    If *this\msg\src
      PokeS(*this\msg\src, src, len, #PB_Ascii)
      ProcedureReturn #True
    Else
      ProcedureReturn #False
    EndIf
  EndIf
  
  ProcedureReturn #True
EndProcedure
Procedure.b ____AWPB_Message_SetMailSourceBin(*this.____AWPB_Message, *src, srclen.i)                                       ; Mailquellcode setzen; es wird ein Binrer ASCII-String bergeben
  ____AWPB_Message_Clear(*this)
  If *src And srclen
    *this\msg\decoded = #False
    *this\msg\src     = AllocateMemory(srclen + 1)
    If *this\msg\src
      CopyMemory(*src, *this\msg\src, srclen)
      PokeA(*this\msg\src + srclen, 0)
      ProcedureReturn #True
    Else
      ProcedureReturn #False
    EndIf
  EndIf
  
  ProcedureReturn #True
EndProcedure
Procedure.s ____AWPB_Message_GetMailSource(*this.____AWPB_Message)                                                          ; Mailquellcode zurckgeben. es wird ein String zurckgegeben
  If *this\msg\src
    ProcedureReturn PeekS(*this\msg\src, MemorySize(*this\msg\src), #PB_Ascii)
  EndIf
  
  ProcedureReturn ""
EndProcedure
Procedure.i ____AWPB_Message_GetMailSourceBin(*this.____AWPB_Message)                                                       ; Mailquellcode zurckgeben. es wird ein Zeiger auf die Originaldaten zurckgegeben (NICHT FREIGEBEN)
  ProcedureReturn *this\msg\src
EndProcedure
Procedure.b ____AWPB_Message_LoadEML(*this.____AWPB_Message, filename.s)                                                    ; eine Mail im Plain-Text-Format lesen (EML, ASCII); OHNE bcc daten
  Protected file.i, length.i, bytes.i
  
  ____AWPB_Message_Clear(*this)
  If filename
    *this\msg\decoded = #False
    file              = ReadFile(#PB_Any, filename)
    If IsFile(file)
      length = Lof(file)
      If length > 0
        *this\msg\src = AllocateMemory(length + 1)
        If *this\msg\src
          bytes = ReadData(file, *this\msg\src, length)
          If bytes = length
            PokeA(*this\msg\src + length, 0)
          Else
            FreeMemory(*this\msg\src)
            *this\msg\src = #Null          
          EndIf
        EndIf
      EndIf
      CloseFile(file)
    EndIf   
  EndIf
  
  If *this\msg\src : ProcedureReturn #True : EndIf
  ProcedureReturn #False
EndProcedure
Procedure.b ____AWPB_Message_SaveEML(*this.____AWPB_Message, filename.s)                                                    ; eine Mail im Plain-Text-Format screiben (EML, ASCII); OHNE bcc daten
  Protected file.i
  
  If *this\msg\src
    file = CreateFile(#PB_Any, filename)
    If IsFile(file)
      WriteData(file, *this\msg\src, MemorySize(*this\msg\src) - 1)   ; im Memorysize ist das abschlieende Chr(0) drin, das weglasen
      CloseFile(file)
      ProcedureReturn #True
    EndIf
  EndIf
  
  ProcedureReturn #False
EndProcedure
Procedure.b ____AWPB_Message_Load(*this.____AWPB_Message, filename.s)                                                       ; eine Mail im binren format laden; MIT bcc daten
  Protected *f.AWPB_ChunkyFile, res.b, hdr.s
  
  ____AWPB_Message_Clear(*this)
  *this\msg\decoded = #False
  res               = #False
  If filename
    *f = AWPB_CreateChunkyFileObj()
    If *f
      If *f\Open(FileName, "MAIL", #AWPB_FILE_MODE_READ)
        While *f\NextChunk()
;           If Not *f\CheckChunkIntegrity()
;             res = #False
;             Break
;           EndIf
          hdr = *f\GetChunkHeader()
          If hdr = "BCC "
            *this\msg\bcc = *f\ReadString()
          ElseIf hdr = "SRC "
            If Not *this\msg\src
              *this\msg\src = *f\ReadBlock()
              res = #True
            Else
              res = #False
              Break
            EndIf
          EndIf
        Wend
        *f\Close()
      EndIf
      *f\Destroy()  
    EndIf
  EndIf
  
  ProcedureReturn res
EndProcedure
Procedure.b ____AWPB_Message_Save(*this.____AWPB_Message, filename.s, complevel.i)                                          ; eine Mail im binren format speichern (komprimiert); MIT bcc daten
  Protected *f.AWPB_ChunkyFile, res.b
  
  res = #False
  If *this\msg\src And filename
    *f = AWPB_CreateChunkyFileObj()
    If *f
      If *f\Open(FileName, "MAIL", #AWPB_FILE_MODE_WRITE)
        If *this\msg\bcc
          If *f\OpenChunk("BCC ")
            res = *f\WriteString(*this\msg\bcc, #PB_Ascii, complevel)
            res & *f\CloseChunk()
          EndIf
        Else
          res = #True
        EndIf
        If res
          If *f\OpenChunk("SRC ")
            res = *f\WriteBlock(*this\msg\src, MemorySize(*this\msg\src), complevel)
            res & *f\CloseChunk()
          EndIf
        EndIf
        *f\Close()
      EndIf
      *f\Destroy()
    EndIf
  EndIf
  
  ProcedureReturn res
EndProcedure
Procedure.s ____AWPB_Message_GetHeaderField(*this.____AWPB_Message, field.i, *part.AWPB_MSG_MimePart)
  Protected res.s
  
  If Not *part : *part = *this\msg : EndIf
  Select field 
    Case #AWPB_HDR_SUBJECT:                   res = ____MSG_DecodeInline(____AWPB_Message_GetHeaderData(*this, *part, "SUBJECT"))
    Case #AWPB_HDR_FROM:                      res = ____MSG_DecodeInline(____AWPB_Message_GetHeaderData(*this, *part, "FROM"))
    Case #AWPB_HDR_CC:                        res = ____MSG_DecodeInline(____AWPB_Message_GetHeaderData(*this, *part, "CC"))
    Case #AWPB_HDR_CC_MAILS:                  res = ____MSG_CreateMailAdrList(____MSG_DecodeInline(____AWPB_Message_GetHeaderData(*this, *part, "CC")), ", ")
    Case #AWPB_HDR_TO:                        res = ____MSG_DecodeInline(____AWPB_Message_GetHeaderData(*this, *part, "TO"))
    Case #AWPB_HDR_TO_MAILS:                  res = ____MSG_CreateMailAdrList(____MSG_DecodeInline(____AWPB_Message_GetHeaderData(*this, *part, "TO")), ", ")
    Case #AWPB_HDR_BCC:                       res = *this\msg\bcc
    Case #AWPB_HDR_BCC_MAILS:                 res = ____MSG_CreateMailAdrList(*this\msg\bcc, ", ")
    Case #AWPB_HDR_RETURNPATH:                res = ____MSG_DecodeInline(____AWPB_Message_GetHeaderData(*this, *part, "RETURN-PATH"))
    Case #AWPB_HDR_MESSAGEID:                 res = ____AWPB_Message_GetHeaderData(*this, *part, "MESSAGE-ID")
    Case #AWPB_HDR_DATE:                      res = ____AWPB_Message_GetHeaderData(*this, *part, "DATE")
    Case #AWPB_HDR_DISPOSITIONNOTIFICATIONTO: res = ____MSG_DecodeInline(____AWPB_Message_GetHeaderData(*this, *part, "DISPOSITION-NOTIFICATION-TO"))
    Case #AWPB_HDR_REPLYTO:                   res = ____MSG_DecodeInline(____AWPB_Message_GetHeaderData(*this, *part, "REPLY-TO"))
    Case #AWPB_HDR_CONTENTTYPE:               res = ____AWPB_Message_GetHeaderData(*this, *part, "CONTENT-TYPE") : If res : res = StringField(res, 1, "/") : EndIf
    Case #AWPB_HDR_CONTENTSUBTYPE:            res = ____AWPB_Message_GetHeaderData(*this, *part, "CONTENT-TYPE") : If res : res = StringField(res, 2, "/") : EndIf
    Case #AWPB_HDR_CONTENTCHARSET:            res = ____AWPB_Message_GetHeaderValue(*this, *part, "CONTENT-TYPE", "CHARSET")
    Case #AWPB_HDR_CONTENTTRANSFERENCODING:   res = ____AWPB_Message_GetHeaderData(*this, *part, "CONTENT-TRANSFER-ENCODING")
    Case #AWPB_HDR_BOUNDARY:                  res = ____AWPB_Message_GetHeaderValue(*this, *part, "CONTENT-TYPE", "BOUNDARY")
    Case #AWPB_HDR_CONTENTDISPOSITION:        res = ____AWPB_Message_GetHeaderData(*this, *part, "CONTENT-DISPOSITION")
    Case #AWPB_HDR_CONTENTNAME:               res = ____AWPB_Message_GetHeaderValue(*this, *part, "CONTENT-TYPE", "NAME")
    Case #AWPB_HDR_CONTENTFILENAME:           res = ____AWPB_Message_GetHeaderValue(*this, *part, "CONTENT-DISPOSITION", "FILENAME")
    Case #AWPB_HDR_CONTENTID:                 res = ____AWPB_Message_GetHeaderData(*this, *part, "CONTENT-ID")
    Case #AWPB_HDR_XPRIORITY:                 res = ____AWPB_Message_GetHeaderData(*this, *part, "X-PRIORITY")
    Default:                                  res = ""
  EndSelect
  
  ProcedureReturn res
EndProcedure
Procedure   ____AWPB_Message_SetHeaderField(*this.____AWPB_Message, field.i, value.s, *part.AWPB_MSG_MimePart)
  If Not *part : *part = *this\msg : EndIf
  Select field 
    Case #AWPB_HDR_SUBJECT:                   ____AWPB_Message_AddHeader(*this, *part, "Subject", ____MSG_EncodeInline(value, "", #AWPB_MSG_ENCODE_AUTO, #False, #True)) 
    Case #AWPB_HDR_FROM:                      ____AWPB_Message_AddHeader(*this, *part, "From", value)
    Case #AWPB_HDR_CC, #AWPB_HDR_CC_MAILS:    ____AWPB_Message_AddHeader(*this, *part, "CC", value) 
    Case #AWPB_HDR_TO, #AWPB_HDR_TO_MAILS:    ____AWPB_Message_AddHeader(*this, *part, "To", value) 
    Case #AWPB_HDR_BCC, #AWPB_HDR_BCC_MAILS:  *this\msg\bcc = value
    Case #AWPB_HDR_DATE:                      ____AWPB_Message_AddHeader(*this, *part, "Date", value) 
    Case #AWPB_HDR_DISPOSITIONNOTIFICATIONTO: ____AWPB_Message_AddHeader(*this, *part, "Disposition-Notification-To", value) 
    Case #AWPB_HDR_REPLYTO:                   ____AWPB_Message_AddHeader(*this, *part, "Reply-To", value) 
    Case #AWPB_HDR_CONTENTCHARSET:            ____AWPB_Message_AddHeaderValue(*this, *part, "Content-Type", "charset", value) 
    Case #AWPB_HDR_CONTENTTRANSFERENCODING:   ____AWPB_Message_AddHeader(*this, *part, "Content-Transfer-Encoding", value) 
    Case #AWPB_HDR_BOUNDARY:                  ____AWPB_Message_AddHeaderValue(*this, *part, "Content-Type", "boundary", ____MSG_AddQuotesIfNeeded(value))
    Case #AWPB_HDR_CONTENTDISPOSITION:        ____AWPB_Message_AddHeader(*this, *part, "Content-Disposition", value) 
    Case #AWPB_HDR_CONTENTNAME:               ____AWPB_Message_AddHeaderValue(*this, *part, "Content-Type", "name", value) 
    Case #AWPB_HDR_CONTENTFILENAME:           ____AWPB_Message_AddHeaderValue(*this, *part, "Content-Disposition", "filename", value) 
    Case #AWPB_HDR_CONTENTID:                 ____AWPB_Message_AddHeader(*this, *part, "Content-ID", value) 
    Case #AWPB_HDR_XPRIORITY:                 ____AWPB_Message_AddHeader(*this, *part, "X-Priority", value) 
    Case #AWPB_HDR_FULLCONTENTTYPE:           ____AWPB_Message_AddHeader(*this, *part, "Content-Type", value) 
  EndSelect
EndProcedure
Procedure   ____AWPB_Message_SetContentType(*this.____AWPB_Message, contenttype.s, contentsubtype.s, *part.AWPB_MSG_MimePart)
  ____AWPB_Message_SetHeaderField(*this, #AWPB_HDR_FULLCONTENTTYPE, contenttype + "/" + contentsubtype, *part)
EndProcedure
Procedure.i ____AWPB_Message_FindPartCID(*this.____AWPB_Message, cid.s, casesensitive.b, *part.AWPB_MSG_MimePart)
  Protected res.i
  
  If Not *part : *part = *this\msg : EndIf
  If casesensitive  
    If ____AWPB_Message_GetHeaderField(*this, #AWPB_HDR_CONTENTID, *part) = cid
      ProcedureReturn *part
    EndIf
  Else
    If UCase(____AWPB_Message_GetHeaderField(*this, #AWPB_HDR_CONTENTID, *part)) = UCase(cid)
      ProcedureReturn *part
    EndIf
  EndIf      
  If UCase(____AWPB_Message_GetHeaderField(*this, #AWPB_HDR_CONTENTTYPE, *part)) = "MULTIPART"
    ForEach *part\MimePart()
      res = ____AWPB_Message_FindPartCID(*this, cid, casesensitive, *part\MimePart())
      If res : ProcedureReturn res : EndIf
    Next
  EndIf
  
  ProcedureReturn #Null
EndProcedure
Procedure.i ____AWPB_Message_FindPartName(*this.____AWPB_Message, name.s, casesensitive.b, *part.AWPB_MSG_MimePart)
  Protected res.i
  
  If Not *part : *part = *this\msg : EndIf
  If casesensitive
    If ____AWPB_Message_GetHeaderField(*this, #AWPB_HDR_CONTENTNAME, *part) = name
      ProcedureReturn *part
    EndIf
  Else
    If UCase(____AWPB_Message_GetHeaderField(*this, #AWPB_HDR_CONTENTNAME, *part)) = UCase(name)
      ProcedureReturn *part
    EndIf
  EndIf      
  If UCase(____AWPB_Message_GetHeaderField(*this, #AWPB_HDR_CONTENTTYPE, *part)) = "MULTIPART"
    ForEach *part\MimePart()
      res = ____AWPB_Message_FindPartName(*this, name, casesensitive, *part\MimePart())
      If res : ProcedureReturn res : EndIf
    Next
  EndIf
  
  ProcedureReturn #Null
EndProcedure
Procedure.i ____AWPB_Message_FindPartFileName(*this.____AWPB_Message, filename.s, casesensitive.b, *part.AWPB_MSG_MimePart)
  Protected res.i, fn.s
  
  If Not *part : *part = *this\msg : EndIf
  If casesensitive
    If ____AWPB_Message_GetHeaderField(*this, #AWPB_HDR_CONTENTFILENAME, *part) = filename
      ProcedureReturn *part
    EndIf
  Else
    If UCase(____AWPB_Message_GetHeaderField(*this, #AWPB_HDR_CONTENTFILENAME, *part)) = UCase(filename)
      ProcedureReturn *part
    EndIf
  EndIf
  If UCase(____AWPB_Message_GetHeaderField(*this, #AWPB_HDR_CONTENTTYPE, *part)) = "MULTIPART"
    ForEach *part\MimePart()
      res = ____AWPB_Message_FindPartFileName(*this, filename, casesensitive, *part\MimePart())
      If res : ProcedureReturn res : EndIf
    Next
  EndIf
  
  ProcedureReturn #Null
EndProcedure
Procedure.i ____AWPB_Message_GetDate(*this.____AWPB_Message, *part.AWPB_MSG_MimePart)
  ProcedureReturn ____DATE_ParseRFC2822DateTime(____AWPB_Message_GetHeaderField(*this, #AWPB_HDR_DATE, *part))
EndProcedure
Procedure   ____AWPB_Message_SetDate(*this.____AWPB_Message, datevalue.i, *part.AWPB_MSG_MimePart)
  ____AWPB_Message_SetHeaderField(*this, #AWPB_HDR_DATE, ____DATE_GetRFC2822DateTime(datevalue), *part)
EndProcedure
Procedure   ____AWPB_Message_SetDateNow(*this.____AWPB_Message, *part.AWPB_MSG_MimePart)
  ____AWPB_Message_SetHeaderField(*this, #AWPB_HDR_DATE, ____DATE_GetRFC2822DateTime(Date()), *part)
EndProcedure
Procedure.i ____AWPB_Message_GetXPriority(*this.____AWPB_Message, *part.AWPB_MSG_MimePart)
  ProcedureReturn Val(____SUP_TrimWS(____AWPB_Message_GetHeaderField(*this, #AWPB_HDR_XPRIORITY, *part)))
EndProcedure
Procedure   ____AWPB_Message_SetXPriority(*this.____AWPB_Message, priority.i, *part.AWPB_MSG_MimePart)
  ____AWPB_Message_SetHeaderField(*this, #AWPB_HDR_XPRIORITY, Str(priority), *part)
EndProcedure
Procedure.i ____AWPB_Message_GetHTMLPart(*this.____AWPB_Message, *part.AWPB_MSG_MimePart)                                   ; den text/html-Part suchen und zurckgeben
  Protected res.i
  
  If Not *part : *part = *this\msg : EndIf
  If UCase(____AWPB_Message_GetHeaderField(*this, #AWPB_HDR_CONTENTTYPE, *part)) = "TEXT"
    If UCase(____AWPB_Message_GetHeaderField(*this, #AWPB_HDR_CONTENTSUBTYPE, *part)) = "HTML"
      ProcedureReturn *part
    EndIf
  ElseIf UCase(____AWPB_Message_GetHeaderField(*this, #AWPB_HDR_CONTENTTYPE, *part)) = "MULTIPART"
    ForEach *part\MimePart()
      res = ____AWPB_Message_GetHTMLPart(*this, *part\MimePart())
      If res : ProcedureReturn res : EndIf
    Next
  EndIf
  
  ProcedureReturn #Null
EndProcedure
Procedure.i ____AWPB_Message_GetPlainPart(*this.____AWPB_Message, *part.AWPB_MSG_MimePart)                                  ; den text/plain-Part suchen und zurckgeben
  Protected res.i
  
  If Not *part : *part = *this\msg : EndIf
  If UCase(____AWPB_Message_GetHeaderField(*this, #AWPB_HDR_CONTENTTYPE, *part)) = "TEXT"
    If UCase(____AWPB_Message_GetHeaderField(*this, #AWPB_HDR_CONTENTSUBTYPE, *part)) = "PLAIN"
      ProcedureReturn *part
    EndIf
  ElseIf UCase(____AWPB_Message_GetHeaderField(*this, #AWPB_HDR_CONTENTTYPE, *part)) = "MULTIPART"
    ForEach *part\MimePart()
      res = ____AWPB_Message_GetPlainPart(*this, *part\MimePart())
      If res : ProcedureReturn res : EndIf
    Next
  EndIf
  
  ProcedureReturn #Null
EndProcedure
Procedure.s ____AWPB_Message_GetHTML(*this.____AWPB_Message)                                                                ; den text/html-Part suchen und den Body als String zurckgeben
  Protected *part.AWPB_MSG_MimePart, encoding.s
  
  *part = ____AWPB_Message_GetHTMLPart(*this, #Null)
  If *part
    encoding = UCase(____AWPB_Message_GetHeaderField(*this, #AWPB_HDR_CONTENTTRANSFERENCODING, *part))
    If encoding  = "QUOTED-PRINTABLE"
      ProcedureReturn ____CODE_DecodeQuotedPrintableStr(*part\body, *part\bodylen)
    ElseIf encoding = "BASE64"
      ProcedureReturn ____CODE_DecodeBase64Str(*part\body, *part\bodylen)
    Else
      ProcedureReturn PeekS(*part\body, *part\bodylen, #PB_Ascii)
    EndIf
  EndIf
  
  ProcedureReturn ""
EndProcedure
Procedure.s ____AWPB_Message_GetPlain(*this.____AWPB_Message)                                                               ; den text/plain-Part suchen und den Body als String zurckgeben
  Protected *part.AWPB_MSG_MimePart, encoding.s
  
  *part = ____AWPB_Message_GetPlainPart(*this, #Null)
  If *part
    encoding = UCase(____AWPB_Message_GetHeaderField(*this, #AWPB_HDR_CONTENTTRANSFERENCODING, *part))
    If encoding = "QUOTED-PRINTABLE"
      ProcedureReturn ____CODE_DecodeQuotedPrintableStr(*part\body, *part\bodylen)
    ElseIf encoding = "BASE64"
      ProcedureReturn ____CODE_DecodeBase64Str(*part\body, *part\bodylen)
    Else
      ProcedureReturn PeekS(*part\body, *part\bodylen, #PB_Ascii)
    EndIf
  EndIf
  
  ProcedureReturn ""
EndProcedure
Procedure.s ____AWPB_Message_GetPartAsString(*this.____AWPB_Message, *part.AWPB_MSG_MimePart)                               ; den Body des bergebenen Parts auslesen (ggf dekodieren) und als String zurckgeben
  Protected encoding.s
  
  If Not *part : *part = *this\msg : EndIf
  encoding = UCase(____AWPB_Message_GetHeaderField(*this, #AWPB_HDR_CONTENTTRANSFERENCODING, *part))
  If encoding = "QUOTED-PRINTABLE"
    ProcedureReturn ____CODE_DecodeQuotedPrintableStr(*part\body, *part\bodylen)
  ElseIf encoding = "BASE64"
    ProcedureReturn ____CODE_DecodeBase64Str(*part\body, *part\bodylen)
  Else
    ProcedureReturn PeekS(*part\body, *part\bodylen, #PB_Ascii)
  EndIf
EndProcedure
Procedure.i ____AWPB_Message_GetPartAsBin(*this.____AWPB_Message, *part.AWPB_MSG_MimePart, *partlen)                        ; den Body des bergebenen Parts auslesen (ggf dekodieren) und als Zeiger auf Speicheradresse zurckgeben; die Lnge wird in PartLen gespeichert
  Protected *res, len.i, encoding.s
  
  *res = #Null  
  If Not *part : *part = *this\msg : EndIf
  encoding = UCase(____AWPB_Message_GetHeaderField(*this, #AWPB_HDR_CONTENTTRANSFERENCODING, *part))
  If encoding = "QUOTED-PRINTABLE"
    *res = ____CODE_DecodeQuotedPrintable(*part\body, *part\bodylen, @len)
  ElseIf encoding = "BASE64"
    *res = ____CODE_DecodeBase64(*part\body, *part\bodylen, @len)
  Else
    len  = *part\bodylen
    *res = AllocateMemory(len)
    If *res : CopyMemory(*part\body, *res, len) : EndIf
  EndIf
  If *res
    PokeI(*partlen, len)
  Else
    PokeI(*partlen, 0)
  EndIf
  
  ProcedureReturn *res
EndProcedure
Procedure.b ____AWPB_Message_SavePartToFile(*this.____AWPB_Message, *part.AWPB_MSG_MimePart, filename.s)                    ; den Body des bergebenen Parts in die angegebene Date (inkl, Pfad) speichern (ggf vorher dekodieren)
  Protected *bin, len.i, res.b, file.i
  
  If filename
    res  = #False
    *bin = ____AWPB_Message_GetPartAsBin(*this, *part, @len)
    If *bin
      file = CreateFile(#PB_Any, filename)
      If IsFile(file)
        WriteData(file, *bin, len)
        CloseFile(file)
        res = #True
      EndIf
      FreeMemory(*bin)  
    EndIf
  EndIf
  
  ProcedureReturn res
EndProcedure
Procedure.s ____AWPB_Message_GetPlainAsHTML(*this.____AWPB_Message)                                                         ; den text/plain-Part suchen und den Body als in HTML umgewandelten String zurckgeben
  Protected plain.s, html.s, *part.AWPB_MSG_MimePart
  
  html  = ""
  *part = ____AWPB_Message_GetPlainPart(*this, #Null)
  If *part
    plain = ____AWPB_Message_GetPartAsString(*this, *part)
    If plain
      html = "<!DOCTYPE html PUBLIC "+Chr(34)+"-//W3C//DTD html 4.01 Transitional//EN"+Chr(34)+">"+Chr(10)
      html + "<html><head><meta content="+Chr(34)+"text/html; charset=" + html + ____AWPB_Message_GetHeaderField(*this, #AWPB_HDR_CONTENTCHARSET, *part) + Chr(34)
      html + " http-equiv="+Chr(34)+"content-type"+Chr(34)+">"
      html + "<title></title></head><body><pre><code>"
      html + plain
      html + "</code></pre></body></html>"
    EndIf
  EndIf
  
  ProcedureReturn html
EndProcedure
Procedure.b ____AWPB_Message_IsDecoded(*this.____AWPB_Message)
  ProcedureReturn *this\msg\decoded
EndProcedure
Procedure.i ____AWPB_Message_GetSize(*this.____AWPB_Message)
  If *this\msg\src
    ProcedureReturn (MemorySize(*this\msg\src) - 1)
  EndIf
  
  ProcedureReturn 0
EndProcedure
Procedure.i ____AWPB_Message_Destroy(*this.____AWPB_Message)
  ____AWPB_Message_Clear(*this)
  FreeMemory(*this\msg)
  FreeMemory(*this)  
  ProcedureReturn #Null
EndProcedure
;}
DataSection ;{ Virtual Function Table
____AWPB_Message_VT:
  ; public
  Data.i @____AWPB_Message_Destroy()
  Data.i @____AWPB_Message_Clear()
  Data.i @____AWPB_Message_Decode()  
  Data.i @____AWPB_Message_BuildAttachmentList()
  Data.i @____AWPB_Message_FreeAttachmentList()
  Data.i @____AWPB_Message_Encode()  
  Data.i @____AWPB_Message_SetBodyBinary()
  Data.i @____AWPB_Message_SetBodyText()
  Data.i @____AWPB_Message_SetMultipart()
  Data.i @____AWPB_Message_AddMultipart()
  Data.i @____AWPB_Message_AddPartHTML()
  Data.i @____AWPB_Message_AddPartText()
  Data.i @____AWPB_Message_AddPartBinary()
  Data.i @____AWPB_Message_AddPartBinaryFile()
  Data.i @____AWPB_Message_SetMailSource()
  Data.i @____AWPB_Message_SetMailSourceBin()
  Data.i @____AWPB_Message_GetMailSource()
  Data.i @____AWPB_Message_LoadEML()
  Data.i @____AWPB_Message_SaveEML()
  Data.i @____AWPB_Message_Load()
  Data.i @____AWPB_Message_Save()
  Data.i @____AWPB_Message_GetHeaderField()
  Data.i @____AWPB_Message_SetHeaderField()
  Data.i @____AWPB_Message_SetContentType()
  Data.i @____AWPB_Message_FindPartCID()
  Data.i @____AWPB_Message_FindPartName()
  Data.i @____AWPB_Message_FindPartFileName()
  Data.i @____AWPB_Message_GetDate()
  Data.i @____AWPB_Message_SetDate()
  Data.i @____AWPB_Message_SetDateNow()
  Data.i @____AWPB_Message_GetXPriority()
  Data.i @____AWPB_Message_SetXPriority()
  Data.i @____AWPB_Message_GetHTML()
  Data.i @____AWPB_Message_GetPlain()
  Data.i @____AWPB_Message_GetPartAsString()
  Data.i @____AWPB_Message_GetPartAsBin()
  Data.i @____AWPB_Message_SavePartToFile()
  Data.i @____AWPB_Message_GetPlainAsHTML()
  Data.i @____AWPB_Message_GetSize()
  
  ; private
  Data.i @____AWPB_Message_ClearPart()
  Data.i @____AWPB_Message_GetHeaderValue()
  Data.i @____AWPB_Message_GetHeaderData()
  Data.i @____AWPB_Message_AnalyzeMimePart()
  Data.i @____AWPB_Message_AddHeader()
  Data.i @____AWPB_Message_AddHeaderValue()  
  Data.i @____AWPB_Message_BuildHeader()
  Data.i @____AWPB_Message_CalcHeaderSize()
  Data.i @____AWPB_Message_CalcSize()
  Data.i @____AWPB_Message_BuildPart()
  Data.i @____AWPB_Message_GetHTMLPart()
  Data.i @____AWPB_Message_GetPlainPart()
  Data.i @____AWPB_Message_GetMailSourceBin() 
  Data.i @____AWPB_Message_IsDecoded()
EndDataSection ;}
;}
 

